@@ -1,5 +1,6 @@
-use crate::block::ID;
+use crate::block::{Block, ID};
 use crate::block_store::BlockStore;
+use crate::store::Store;
 use crate::updates::decoder::{Decode, Decoder};
 use crate::updates::encoder::{Encode, Encoder};
 use crate::utils::client_hasher::ClientHasher;
@@ -95,7 +96,11 @@ impl IdRange {
         match self {
             IdRange::Continuous(r) => {
                 if r.end >= range.start {
-                    r.end = range.end; // two ranges overlap, we can eagerly merge them
+                    if r.start > range.end {
+                        *self = IdRange::Fragmented(vec![range, r.clone()])
+                    } else {
+                        r.end = range.end; // two ranges overlap, we can eagerly merge them
+                    }
                 } else {
                     *self = IdRange::Fragmented(vec![r.clone(), range])
                 }
@@ -148,6 +153,39 @@ impl IdRange {
             }
         }
     }
+
+    fn merge(&mut self, other: IdRange) {
+        let mut raw = std::mem::take(self);
+        *self = match (raw, other) {
+            (IdRange::Continuous(mut a), IdRange::Continuous(b)) => {
+                if a.end >= b.start && a.start <= b.start {
+                    a.end = b.end;
+                    IdRange::Continuous(a)
+                } else {
+                    IdRange::Fragmented(vec![a, b])
+                }
+            }
+            (IdRange::Fragmented(mut a), IdRange::Continuous(b)) => {
+                a.push(b);
+                IdRange::Fragmented(a)
+            }
+            (IdRange::Continuous(a), IdRange::Fragmented(b)) => {
+                let mut v = b;
+                v.push(a.clone());
+                IdRange::Fragmented(v)
+            }
+            (IdRange::Fragmented(mut a), IdRange::Fragmented(mut b)) => {
+                a.append(&mut b);
+                IdRange::Fragmented(a)
+            }
+        };
+    }
+}
+
+impl Default for IdRange {
+    fn default() -> Self {
+        IdRange::Continuous(0..0)
+    }
 }
 
 impl Encode for IdRange {
@@ -281,35 +319,10 @@ impl IdSet {
     /// ranges and squashing them if necessary.
     pub fn merge(&mut self, other: Self) {
         other.0.into_iter().for_each(|(client, range)| {
-            match self.0.entry(client) {
-                Entry::Occupied(mut e) => {
-                    let r = e.get_mut();
-                    match (r, range) {
-                        (IdRange::Continuous(r1), IdRange::Continuous(r2)) => {
-                            if r1.end >= r2.start && r1.start <= r2.start {
-                                r1.end = r2.end;
-                            } else {
-                                let new = IdRange::Fragmented(vec![r1.clone(), r2.clone()]);
-                                e.replace_entry(new);
-                            }
-                        }
-                        (IdRange::Fragmented(rs), IdRange::Continuous(r)) => {
-                            rs.push(r);
-                        }
-                        (IdRange::Continuous(r), IdRange::Fragmented(rs)) => {
-                            let mut v = rs;
-                            v.push(r.clone());
-                            let new = IdRange::Fragmented(v);
-                            e.replace_entry(new);
-                        }
-                        (IdRange::Fragmented(rs1), IdRange::Fragmented(mut rs2)) => {
-                            rs1.append(&mut rs2);
-                        }
-                    }
-                }
-                Entry::Vacant(e) => {
-                    e.insert(range);
-                }
+            if let Some(r) = self.0.get_mut(&client) {
+                r.merge(range)
+            } else {
+                self.0.insert(client, range);
             }
         });
         self.squash()
@@ -454,8 +467,33 @@ impl DeleteSet {
         self.0.squash()
     }
 
-    pub(crate) fn try_compact(&mut self, blocks: &BlockStore) {
-        //TODO
+    pub(crate) fn try_squash_with(&mut self, store: &mut Store) {
+        // try to merge deleted / gc'd items
+        for (client, range) in self.iter() {
+            if let Some(mut blocks) = store.blocks.get_mut(client) {
+                for r in range.iter().rev() {
+                    // start with merging the item next to the last deleted item
+                    let mut si = (blocks.len() - 1)
+                        .min(1 + blocks.find_pivot(r.end - 1).unwrap_or_default());
+                    let mut block = &blocks[si];
+                    while si > 0 && block.id().clock >= r.start {
+                        if let Some(compaction) = blocks.squash_left(si) {
+                            if let Some(right) = compaction.new_right {
+                                right.fix_pivot((right.pivot().max(1) - 1) as u32);
+                                if let Block::Item(item) =
+                                    store.blocks.get_block_mut(&right).unwrap()
+                                {
+                                    item.left = Some(compaction.replacement);
+                                }
+                                blocks = store.blocks.get_mut(client).unwrap();
+                            }
+                        }
+                        si -= 1;
+                        block = &blocks[si];
+                    }
+                }
+            }
+        }
     }
 }
 
