@@ -1,5 +1,5 @@
 use crate::block::{BlockPtr, ItemContent, ItemPosition, Prelim};
-use crate::types::{Branch, BranchRef, TypePtr, Value, TYPE_REFS_ARRAY};
+use crate::types::{Branch, BranchRef, Event, Observer, TypePtr, Value, TYPE_REFS_ARRAY};
 use crate::Transaction;
 use lib0::any::Any;
 use std::collections::VecDeque;
@@ -105,13 +105,13 @@ impl Array {
     /// of the range of a current array.
     pub fn get(&self, txn: &Transaction, index: u32) -> Option<Value> {
         let inner = self.0.borrow();
-        let (content, idx) = inner.get_at(txn, index)?;
+        let (content, idx) = inner.get_at(&txn.store().blocks, index)?;
         Some(content.get_content(txn).remove(idx))
     }
 
     /// Returns an iterator, that can be used to lazely traverse over all values stored in a current
     /// array.
-    pub fn iter<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> ArrayIter<'b, 'txn> {
+    pub fn iter<'a, 'b>(&'a self, txn: &'b Transaction) -> ArrayIter<'b> {
         ArrayIter::new(self, txn)
     }
 
@@ -120,16 +120,31 @@ impl Array {
         let res = self.iter(txn).map(|v| v.to_json(txn)).collect();
         Any::Array(res)
     }
+
+    /// Subscribes a given callback to be triggered whenever current array is changed.
+    /// A callback is triggered whenever a transaction gets committed. This function does not
+    /// trigger if changes have been observed by nested shared collections.
+    ///
+    /// All array changes can be tracked by using [Event::delta] method.
+    ///
+    /// Returns an [Observer] which, when dropped, will unsubscribe current callback.
+    pub fn observe<F>(&self, f: F) -> Observer
+    where
+        F: Fn(&Transaction, &Event) -> () + 'static,
+    {
+        let mut branch = self.0.borrow_mut();
+        branch.observe(f)
+    }
 }
 
-pub struct ArrayIter<'b, 'txn> {
+pub struct ArrayIter<'b> {
     content: VecDeque<Value>,
     ptr: Option<BlockPtr>,
-    txn: &'b Transaction<'txn>,
+    txn: &'b Transaction,
 }
 
-impl<'b, 'txn> ArrayIter<'b, 'txn> {
-    fn new(array: &Array, txn: &'b Transaction<'txn>) -> Self {
+impl<'b> ArrayIter<'b> {
+    fn new(array: &Array, txn: &'b Transaction) -> Self {
         let inner = array.0.borrow();
         ArrayIter {
             ptr: inner.start,
@@ -139,14 +154,14 @@ impl<'b, 'txn> ArrayIter<'b, 'txn> {
     }
 }
 
-impl<'b, 'txn> Iterator for ArrayIter<'b, 'txn> {
+impl<'b> Iterator for ArrayIter<'b> {
     type Item = Value;
 
     fn next(&mut self) -> Option<Self::Item> {
         match self.content.pop_front() {
             None => {
                 if let Some(ptr) = self.ptr.take() {
-                    let item = self.txn.store.blocks.get_item(&ptr)?;
+                    let item = self.txn.store().blocks.get_item(&ptr)?;
                     self.ptr = item.right.clone();
                     if !item.is_deleted() && item.is_countable() {
                         self.content = item.content.get_content(self.txn).into();
@@ -194,12 +209,12 @@ where
     T: IntoIterator<Item = V>,
     V: Into<Any>,
 {
-    fn into_content(self, txn: &mut Transaction, ptr: TypePtr) -> (ItemContent, Option<Self>) {
+    fn into_content(self, _txn: &mut Transaction, _ptr: TypePtr) -> (ItemContent, Option<Self>) {
         let vec: Vec<Any> = self.0.into_iter().map(|v| v.into()).collect();
         (ItemContent::Any(vec), None)
     }
 
-    fn integrate(self, txn: &mut Transaction, inner_ref: BranchRef) {}
+    fn integrate(self, _txn: &mut Transaction, _inner_ref: BranchRef) {}
 }
 
 impl<T, V> Prelim for PrelimArray<T, V>
@@ -207,7 +222,7 @@ where
     V: Prelim,
     T: IntoIterator<Item = V>,
 {
-    fn into_content(self, txn: &mut Transaction, ptr: TypePtr) -> (ItemContent, Option<Self>) {
+    fn into_content(self, _txn: &mut Transaction, ptr: TypePtr) -> (ItemContent, Option<Self>) {
         let inner = BranchRef::new(Branch::new(ptr, TYPE_REFS_ARRAY, None));
         (ItemContent::Type(inner), Some(self))
     }
@@ -222,12 +237,16 @@ where
 
 #[cfg(test)]
 mod test {
-    use crate::test_utils::exchange_updates;
+    use crate::test_utils::{exchange_updates, run_scenario, RngExt};
     use crate::types::map::PrelimMap;
-    use crate::types::Value;
-    use crate::Doc;
+    use crate::types::{Change, Value};
+    use crate::{Doc, PrelimArray, Update, ID};
     use lib0::any::Any;
-    use std::collections::HashMap;
+    use rand::prelude::StdRng;
+    use rand::Rng;
+    use std::cell::{Cell, RefCell};
+    use std::collections::{HashMap, HashSet};
+    use std::rc::Rc;
 
     #[test]
     fn push_back() {
@@ -564,11 +583,299 @@ mod test {
             let mut expected = HashMap::new();
             expected.insert("value".to_owned(), Any::Number(i as f64));
             match value {
-                Value::YMap(ref map) => {
-                    assert_eq!(value.to_json(&txn), Any::Map(expected))
+                Value::YMap(_) => {
+                    assert_eq!(value.to_json(&txn), Any::Map(Box::new(expected)))
                 }
-                other => panic!("Value of array at index {} was no YMap", i),
+                _ => panic!("Value of array at index {} was no YMap", i),
             }
         }
     }
+
+    #[test]
+    fn insert_and_remove_events() {
+        let d = Doc::with_client_id(1);
+        let array = {
+            let mut txn = d.transact();
+            txn.get_array("array")
+        };
+        let happened = Rc::new(Cell::new(false));
+        let happened_clone = happened.clone();
+        let _sub = array.observe(move |_, _| {
+            happened_clone.set(true);
+        });
+
+        {
+            let mut txn = d.transact();
+            array.insert_range(&mut txn, 0, [0, 1, 2]);
+            // txn is committed at the end of this scope
+        }
+        assert!(
+            happened.replace(false),
+            "insert of [0,1,2] should trigger event"
+        );
+
+        {
+            let mut txn = d.transact();
+            array.remove_range(&mut txn, 0, 1);
+            // txn is committed at the end of this scope
+        }
+        assert!(
+            happened.replace(false),
+            "removal of [0] should trigger event"
+        );
+
+        {
+            let mut txn = d.transact();
+            array.remove_range(&mut txn, 0, 2);
+            // txn is committed at the end of this scope
+        }
+        assert!(
+            happened.replace(false),
+            "removal of [1,2] should trigger event"
+        );
+    }
+
+    #[test]
+    fn insert_and_remove_event_changes() {
+        let d1 = Doc::with_client_id(1);
+        let array = {
+            let mut txn = d1.transact();
+            txn.get_array("array")
+        };
+        let added = Rc::new(RefCell::new(None));
+        let removed = Rc::new(RefCell::new(None));
+        let delta = Rc::new(RefCell::new(None));
+
+        let (added_c, removed_c, delta_c) = (added.clone(), removed.clone(), delta.clone());
+        let _sub = array.observe(move |txn, e| {
+            *added_c.borrow_mut() = Some(e.inserts(txn).clone());
+            *removed_c.borrow_mut() = Some(e.removes(txn).clone());
+            *delta_c.borrow_mut() = Some(e.delta(txn).to_vec());
+        });
+
+        {
+            let mut txn = d1.transact();
+            array.push_back(&mut txn, 4);
+            array.push_back(&mut txn, "dtrn");
+            // txn is committed at the end of this scope
+        }
+        assert_eq!(
+            added.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 0), ID::new(1, 1)]))
+        );
+        assert_eq!(removed.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Added(vec![
+                Any::Number(4.0).into(),
+                Any::String("dtrn".into()).into()
+            ])])
+        );
+
+        {
+            let mut txn = d1.transact();
+            array.remove_range(&mut txn, 0, 1);
+        }
+        assert_eq!(added.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            removed.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 0)]))
+        );
+        assert_eq!(delta.borrow_mut().take(), Some(vec![Change::Removed(1)]));
+
+        {
+            let mut txn = d1.transact();
+            array.insert(&mut txn, 1, 0.5);
+        }
+        assert_eq!(
+            added.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 2)]))
+        );
+        assert_eq!(removed.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![
+                Change::Retain(1),
+                Change::Added(vec![Any::Number(0.5).into()])
+            ])
+        );
+
+        let d2 = Doc::with_client_id(2);
+        let array2 = {
+            let mut txn = d2.transact();
+            txn.get_array("array")
+        };
+        let (added_c, removed_c, delta_c) = (added.clone(), removed.clone(), delta.clone());
+        let _sub = array2.observe(move |txn, e| {
+            *added_c.borrow_mut() = Some(e.inserts(txn).clone());
+            *removed_c.borrow_mut() = Some(e.removes(txn).clone());
+            *delta_c.borrow_mut() = Some(e.delta(txn).to_vec());
+        });
+
+        {
+            let t1 = d1.transact();
+            let mut t2 = d2.transact();
+
+            let sv = t2.state_vector();
+            let mut encoder = EncoderV1::new();
+            t1.encode_diff(&sv, &mut encoder);
+            t2.apply_update(Update::decode_v1(encoder.to_vec().as_slice()));
+        }
+
+        assert_eq!(
+            added.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 1)]))
+        );
+        assert_eq!(removed.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Added(vec![
+                Any::String("dtrn".into()).into(),
+                Any::Number(0.5).into(),
+            ])])
+        );
+    }
+
+    #[test]
+    fn target_on_local_and_remote() {
+        let d1 = Doc::with_client_id(1);
+        let d2 = Doc::with_client_id(2);
+        let a1 = {
+            let mut txn = d1.transact();
+            txn.get_array("array")
+        };
+        let a2 = {
+            let mut txn = d2.transact();
+            txn.get_array("array")
+        };
+
+        let c1 = Rc::new(RefCell::new(None));
+        let c1c = c1.clone();
+        let _s1 = a1.observe(move |_, e| {
+            *c1c.borrow_mut() = Some(e.target());
+        });
+        let c2 = Rc::new(RefCell::new(None));
+        let c2c = c2.clone();
+        let _s2 = a2.observe(move |_, e| {
+            *c2c.borrow_mut() = Some(e.target());
+        });
+
+        {
+            let mut t1 = d1.transact();
+            a1.insert_range(&mut t1, 0, [1, 2]);
+        }
+        exchange_updates(&[&d1, &d2]);
+
+        assert_eq!(c1.borrow_mut().take(), Some(Value::YArray(a1)));
+        assert_eq!(c2.borrow_mut().take(), Some(Value::YArray(a2)));
+    }
+
+    use crate::updates::decoder::Decode;
+    use crate::updates::encoder::{Encoder, EncoderV1};
+    use std::sync::atomic::{AtomicI64, Ordering};
+
+    static UNIQUE_NUMBER: AtomicI64 = AtomicI64::new(0);
+
+    fn get_unique_number() -> i64 {
+        UNIQUE_NUMBER.fetch_add(1, Ordering::SeqCst)
+    }
+
+    fn array_transactions() -> [Box<dyn Fn(&mut Doc, &mut StdRng)>; 4] {
+        fn insert(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let yarray = txn.get_array("array");
+            let unique_number = get_unique_number();
+            let len = rng.between(1, 4);
+            let content: Vec<_> = (0..len)
+                .into_iter()
+                .map(|_| Any::BigInt(unique_number))
+                .collect();
+            let mut pos = rng.between(0, yarray.len()) as usize;
+            if let Any::Array(expected) = yarray.to_json(&txn) {
+                let mut expected = Vec::from(expected);
+                yarray.insert_range(&mut txn, pos as u32, content.clone());
+
+                for any in content {
+                    expected.insert(pos, any);
+                    pos += 1;
+                }
+                let actual = yarray.to_json(&txn);
+                assert_eq!(actual, Any::Array(expected.into_boxed_slice()))
+            } else {
+                panic!("should not happen")
+            }
+        }
+
+        fn insert_type_array(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let yarray = txn.get_array("array");
+            let pos = rng.between(0, yarray.len());
+            yarray.insert(&mut txn, pos, PrelimArray::from([1, 2, 3, 4]));
+            if let Value::YArray(array2) = yarray.get(&txn, pos).unwrap() {
+                let expected: Box<[Any]> = (1..=4).map(|i| Any::Number(i as f64)).collect();
+                assert_eq!(array2.to_json(&txn), Any::Array(expected));
+            } else {
+                panic!("should not happen")
+            }
+        }
+
+        fn insert_type_map(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let yarray = txn.get_array("array");
+            let pos = rng.between(0, yarray.len());
+            yarray.insert(&mut txn, pos, PrelimMap::<i32>::from(HashMap::default()));
+            if let Value::YMap(map) = yarray.get(&txn, pos).unwrap() {
+                map.insert(&mut txn, "someprop".to_string(), 42);
+                map.insert(&mut txn, "someprop".to_string(), 43);
+                map.insert(&mut txn, "someprop".to_string(), 44);
+            } else {
+                panic!("should not happen: {}", txn.store())
+            }
+        }
+
+        fn delete(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let yarray = txn.get_array("array");
+            let len = yarray.len();
+            if len > 0 {
+                let pos = rng.between(0, len - 1);
+                let del_len = rng.between(1, 2.min(len - pos));
+                if rng.gen_bool(0.5) {
+                    if let Value::YArray(array2) = yarray.get(&txn, pos).unwrap() {
+                        let pos = rng.between(0, array2.len() - 1);
+                        let del_len = rng.between(0, 2.min(array2.len() - pos));
+                        array2.remove_range(&mut txn, pos, del_len);
+                    }
+                } else {
+                    if let Any::Array(old_content) = yarray.to_json(&txn) {
+                        let mut old_content = Vec::from(old_content);
+                        yarray.remove_range(&mut txn, pos, del_len);
+                        old_content.drain(pos as usize..(pos + del_len) as usize);
+                        assert_eq!(
+                            yarray.to_json(&txn),
+                            Any::Array(old_content.into_boxed_slice())
+                        );
+                    } else {
+                        panic!("should not happen")
+                    }
+                }
+            }
+        }
+
+        [
+            Box::new(insert),
+            Box::new(insert_type_array),
+            Box::new(insert_type_map),
+            Box::new(delete),
+        ]
+    }
+
+    fn fuzzy(iterations: usize) {
+        run_scenario(0, &array_transactions(), 5, iterations)
+    }
+
+    #[test]
+    fn fuzzy_test_6() {
+        fuzzy(6)
+    }
 }
