@@ -1,6 +1,6 @@
 use crate::block::{BlockPtr, ItemContent};
 use crate::transaction::Transaction;
-use crate::types::{Branch, BranchRef};
+use crate::types::{Branch, BranchRef, Event, Observer};
 use crate::*;
 use std::cell::Ref;
 
@@ -23,12 +23,13 @@ pub struct Text(BranchRef);
 impl Text {
     /// Converts context of this text data structure into a single string value.
     #[allow(clippy::inherent_to_string)]
-    pub fn to_string(&self, txn: &Transaction<'_>) -> String {
+    pub fn to_string(&self, txn: &Transaction) -> String {
         let inner = self.0.as_ref();
         let mut start = inner.start;
         let mut s = String::new();
+        let store = txn.store();
         while let Some(a) = start.as_ref() {
-            if let Some(item) = txn.store.blocks.get_item(&a) {
+            if let Some(item) = store.blocks.get_item(&a) {
                 if !item.is_deleted() {
                     if let block::ItemContent::String(item_string) = &item.content {
                         s.push_str(item_string);
@@ -44,7 +45,7 @@ impl Text {
 
     /// Returns a number of characters visible in a current text data structure.
     pub fn len(&self) -> u32 {
-        self.0.borrow().len()
+        self.0.borrow().content_len
     }
 
     pub(crate) fn inner(&self) -> Ref<Branch> {
@@ -53,8 +54,8 @@ impl Text {
 
     pub(crate) fn find_position(
         &self,
-        txn: &mut Transaction<'_>,
-        mut count: u32,
+        txn: &mut Transaction,
+        index: u32,
     ) -> Option<block::ItemPosition> {
         let mut pos = {
             let inner = self.0.borrow();
@@ -66,26 +67,37 @@ impl Text {
             }
         };
 
+        let store = txn.store_mut();
+        let encoding = store.options.offset_kind;
+        let mut remaining = index;
         while let Some(right_ptr) = pos.right.as_ref() {
-            if count == 0 {
+            if remaining == 0 {
                 break;
             }
 
-            if let Some(mut right) = txn.store.blocks.get_item(right_ptr) {
+            if let Some(mut right) = store.blocks.get_item(right_ptr) {
                 if !right.is_deleted() {
-                    let mut right_len = right.len();
-                    if count < right_len {
+                    let mut block_len = right.len();
+                    let content_len = right.content_len(encoding);
+                    if remaining < content_len {
                         // split right item
+                        let offset = if let ItemContent::String(str) = &right.content {
+                            str.block_offset(remaining, encoding)
+                        } else {
+                            remaining
+                        };
                         let split_ptr = BlockPtr::new(
-                            ID::new(right.id.client, right.id.clock + count),
+                            ID::new(right.id.client, right.id.clock + offset),
                             right_ptr.pivot() as u32,
                         );
-                        let (_, _) = txn.store.blocks.split_block(&split_ptr);
-                        right = txn.store.blocks.get_item(right_ptr).unwrap();
-                        right_len = right.len();
+                        let (_, _) = store.blocks.split_block(&split_ptr);
+                        right = store.blocks.get_item(right_ptr).unwrap();
+                        block_len = right.len();
+                        remaining = 0;
+                    } else {
+                        remaining -= content_len;
                     }
-                    pos.index += right_len;
-                    count -= right_len;
+                    pos.index += block_len;
                 }
                 pos.left = pos.right.take();
                 pos.right = right.right.clone();
@@ -104,7 +116,7 @@ impl Text {
     /// This method will panic if provided `index` is greater than the length of a current text.
     pub fn insert(&self, tr: &mut Transaction, index: u32, chunk: &str) {
         if let Some(pos) = self.find_position(tr, index) {
-            let value = crate::block::PrelimText(chunk.to_owned());
+            let value = crate::block::PrelimText(chunk.into());
             tr.create_item(&pos, value, None);
         } else {
             panic!("The type or the position doesn't exist!");
@@ -126,6 +138,22 @@ impl Text {
             panic!("Couldn't remove {} elements from an array. Only {} of them were successfully removed.", len, removed);
         }
     }
+
+    /// Subscribes a given callback to be triggered whenever current text is changed.
+    /// A callback is triggered whenever a transaction gets committed. This function does not
+    /// trigger if changes have been observed by nested shared collections.
+    ///
+    /// All text changes can be tracked by using [Event::delta] method: keep in mind that delta
+    /// contains collection of individual characters rather than strings.
+    ///
+    /// Returns an [Observer] which, when dropped, will unsubscribe current callback.
+    pub fn observe<F>(&self, f: F) -> Observer
+    where
+        F: Fn(&Transaction, &Event) -> () + 'static,
+    {
+        let mut branch = self.0.borrow_mut();
+        branch.observe(f)
+    }
 }
 
 impl Into<ItemContent> for Text {
@@ -142,7 +170,16 @@ impl From<BranchRef> for Text {
 
 #[cfg(test)]
 mod test {
-    use crate::Doc;
+    use crate::doc::{OffsetKind, Options};
+    use crate::test_utils::{exchange_updates, run_scenario, RngExt};
+    use crate::types::Change;
+    use crate::updates::decoder::Decode;
+    use crate::updates::encoder::{Encoder, EncoderV1};
+    use crate::{Doc, Update};
+    use lib0::any::Any;
+    use rand::prelude::StdRng;
+    use std::cell::RefCell;
+    use std::rc::Rc;
 
     #[test]
     fn append_single_character_blocks() {
@@ -460,4 +497,178 @@ mod test {
         assert_eq!(a, b);
         assert_eq!(a, "H beautifuld!".to_owned());
     }
+
+    #[test]
+    fn insert_and_remove_event_changes() {
+        let d1 = Doc::with_client_id(1);
+        let txt = {
+            let mut txn = d1.transact();
+            txn.get_text("text")
+        };
+        let delta = Rc::new(RefCell::new(None));
+        let delta_c = delta.clone();
+        let _sub = txt.observe(move |txn, e| {
+            *delta_c.borrow_mut() = Some(e.delta(txn).to_vec());
+        });
+
+        // insert initial string
+        {
+            let mut txn = d1.transact();
+            txt.insert(&mut txn, 0, "abcd");
+        }
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Added(vec![
+                Any::String("a".into()).into(),
+                Any::String("b".into()).into(),
+                Any::String("c".into()).into(),
+                Any::String("d".into()).into()
+            ])])
+        );
+
+        // remove middle
+        {
+            let mut txn = d1.transact();
+            txt.remove_range(&mut txn, 1, 2);
+        }
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Retain(1), Change::Removed(2)])
+        );
+
+        // insert again
+        {
+            let mut txn = d1.transact();
+            txt.insert(&mut txn, 1, "ef");
+        }
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![
+                Change::Retain(1),
+                Change::Added(vec![
+                    Any::String("e".into()).into(),
+                    Any::String("f".into()).into()
+                ])
+            ])
+        );
+
+        // replicate data to another peer
+        let d2 = Doc::with_client_id(2);
+        let txt = {
+            let mut txn = d2.transact();
+            txn.get_text("text")
+        };
+        let delta_c = delta.clone();
+        let _sub = txt.observe(move |txn, e| {
+            *delta_c.borrow_mut() = Some(e.delta(txn).to_vec());
+        });
+
+        {
+            let t1 = d1.transact();
+            let mut t2 = d2.transact();
+
+            let sv = t2.state_vector();
+            let mut encoder = EncoderV1::new();
+            t1.encode_diff(&sv, &mut encoder);
+            t2.apply_update(Update::decode_v1(encoder.to_vec().as_slice()));
+        }
+
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Added(vec![
+                Any::String("a".into()).into(),
+                Any::String("e".into()).into(),
+                Any::String("f".into()).into(),
+                Any::String("d".into()).into()
+            ])])
+        );
+    }
+
+    #[test]
+    fn unicode_support() {
+        let d1 = {
+            let mut options = Options::with_client_id(1);
+            options.offset_kind = OffsetKind::Utf32;
+            Doc::with_options(options)
+        };
+        let txt1 = {
+            let mut txn = d1.transact();
+            txn.get_text("test")
+        };
+
+        let d2 = {
+            let mut options = Options::with_client_id(2);
+            options.offset_kind = OffsetKind::Bytes;
+            Doc::with_options(options)
+        };
+        let txt2 = {
+            let mut txn = d2.transact();
+            txn.get_text("test")
+        };
+
+        {
+            let mut txn = d1.transact();
+
+            txt1.insert(&mut txn, 0, "Zażółć gęślą jaźń");
+            assert_eq!(txt1.to_string(&txn), "Zażółć gęślą jaźń");
+            assert_eq!(txt1.len(), 17);
+        }
+
+        exchange_updates(&[&d1, &d2]);
+
+        {
+            let txn = d2.transact();
+            assert_eq!(txt2.to_string(&txn), "Zażółć gęślą jaźń");
+            assert_eq!(txt2.len(), 26);
+        }
+
+        {
+            let mut txn = d1.transact();
+            txt1.remove_range(&mut txn, 9, 3);
+            txt1.insert(&mut txn, 9, "si");
+
+            assert_eq!(txt1.to_string(&txn), "Zażółć gęsi jaźń");
+            assert_eq!(txt1.len(), 16);
+        }
+
+        exchange_updates(&[&d1, &d2]);
+
+        {
+            let txn = d2.transact();
+            assert_eq!(txt2.to_string(&txn), "Zażółć gęsi jaźń");
+            assert_eq!(txt2.len(), 23);
+        }
+    }
+
+    fn text_transactions() -> [Box<dyn Fn(&mut Doc, &mut StdRng)>; 2] {
+        fn insert_text(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let ytext = txn.get_text("text");
+            let pos = rng.between(0, ytext.len());
+            let word = rng.random_string();
+            ytext.insert(&mut txn, pos, word.as_str());
+        }
+
+        fn delete_text(doc: &mut Doc, rng: &mut StdRng) {
+            let mut txn = doc.transact();
+            let ytext = txn.get_text("text");
+            let len = ytext.len();
+            if len > 0 {
+                let pos = rng.between(0, len - 1);
+                let to_delete = rng.between(2, len - pos);
+                ytext.remove_range(&mut txn, pos, to_delete);
+            }
+        }
+
+        [Box::new(insert_text), Box::new(delete_text)]
+    }
+
+    fn fuzzy(iterations: usize) {
+        run_scenario(0, &text_transactions(), 5, iterations)
+    }
+
+    #[test]
+    fn fuzzy_test_3() {
+        fuzzy(3)
+    }
 }
